#!/usr/bin/env python

import sys
import rospy
import moveit_commander
import moveit_msgs.msg
import geometry_msgs.msg
import copy
import time

JOINT_NAMES = ['joint_s', 'joint_l', 'joint_u', 'joint_r', 'joint_b', 'joint_t']

home_pose = [0.0, -0.4296, -1.0122, 0.0, -0.9522, 0.0]


class RobotManager(object):
    def __init__(self):
        # First initialize `moveit_commander`
        moveit_commander.roscpp_initialize(sys.argv)
        self.robot_commander = moveit_commander.RobotCommander()
        self.move_group = moveit_commander.MoveGroupCommander("manipulator")
        self.display_trajectory_publisher = rospy.Publisher('/move_group/display_planned_path',
                                                            moveit_msgs.msg.DisplayTrajectory,
                                                            queue_size=20)
        self.move_group.set_pose_reference_frame("base_link")
        self.move_group.set_end_effector_link("link_t")

        # User adjustable parameters --------------------------------------
        self.max_velocity_scaling_factor = 1.0
        # self.move_group.set_goal_position_tolerance(0.001)
        # self.move_group.set_goal_joint_tolerance(0.00005)
        self.wait_for_user = False  # if True will wait for the user to press enter before each move is executed
        self.print_joint_angles = False  # if True will print the joint angles of the robot at each move
        self.dwell_time_between_moves = 0.0  # the time in seconds to wait between each move
        # ----------------------------------------------------------------

        self.move_group.set_max_acceleration_scaling_factor(1.0)
        self.move_group.set_max_velocity_scaling_factor(self.max_velocity_scaling_factor)

    def get_joint_angles(self, print_joint_angles=True):
        global JOINT_NAMES
        current_state = self.robot_commander.get_current_state()

        joint_index = 0
        joint_angles = []
        for i, name in enumerate(current_state.joint_state.name):
            if name == JOINT_NAMES[joint_index]:
                joint_angles.append(current_state.joint_state.position[i])
                joint_index = joint_index + 1
                if joint_index >= JOINT_NAMES.__len__():
                    break
        if print_joint_angles:
            rospy.loginfo("[%.5f, %.5f, %.5f, %.5f, %.5f, %.5f]",
                          joint_angles[0], joint_angles[1], joint_angles[2],
                          joint_angles[3], joint_angles[4], joint_angles[5])
        return joint_angles

    @staticmethod
    def diff_joint_angles(j1, j2):
        diff_angles = []
        for i in range(j1.__len__()):
            diff_angles.append(j1[i] - j2[i])

        rospy.loginfo("[%.5f, %.5f, %.5f, %.5f, %.5f, %.5f]",
                      diff_angles[0], diff_angles[1], diff_angles[2], diff_angles[3], diff_angles[4], diff_angles[5])

    def move_arm(self):
        waypoints = []
        waypoints.append(geometry_msgs.msg.Pose())
        last_loop_time = time.time()
        start_time_ = time.time()

        while True:
            self.move_group.set_joint_value_target(home_pose)
            self.move_group.set_start_state_to_current_state()
            plan = self.move_group.plan()
            rospy.loginfo("Planning time to move home: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.go(wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
                rospy.loginfo("Made it to the home position with the following joint angle error:")
                self.diff_joint_angles(home_pose,  self.get_joint_angles(print_joint_angles=False))
            else:
                rospy.logerr("Failed to execute move home")
            start_time_ = time.time()

            rospy.logwarn("Pick and Place cycle time: %s", (time.time() - last_loop_time))
            last_loop_time = time.time()
            time.sleep(self.dwell_time_between_moves)

            # --------------------- approach pick --------------------------------------
            approach_pick_goal = geometry_msgs.msg.Pose()
            approach_pick_goal.orientation.y = 0.707
            approach_pick_goal.orientation.w = 0.707
            approach_pick_goal.position.x = 0.35
            approach_pick_goal.position.y = 0.3
            approach_pick_goal.position.z = 0.4
            # self.move_group.clear_pose_targets()
            # self.move_group.set_pose_target(approach_pick_goal)
            # -------------------
            # This move is hard coded with joint angles to remove the fact that
            # MoveIt does not always find the same final joint pose
            # -------------------
            approach_pick_joint_angles = [0.70802, 0.12343, -0.62547, 0.00162, -0.82092, -0.70912]
            self.move_group.set_joint_value_target(approach_pick_joint_angles)
            self.move_group.set_start_state_to_current_state()

            plan = self.move_group.plan()
            rospy.loginfo("Planning time to move to approach pick: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.go(wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
                rospy.loginfo("Made it to approach pick with the following joint angle error:")
                self.diff_joint_angles(approach_pick_joint_angles,  self.get_joint_angles(print_joint_angles=False))
            else:
                rospy.logerr("Failed to execute approach pick")
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Approach Pick joint angles:")
                self.get_joint_angles()

            # --------------------- pick --------------------------------------
            pick_goal = copy.deepcopy(approach_pick_goal)
            pick_goal.position.z = approach_pick_goal.position.z - 0.2
            waypoints[0] = pick_goal
            self.move_group.set_start_state_to_current_state()
            (plan, fraction) = self.move_group.compute_cartesian_path(waypoints, 0.05, 0)
            plan = self.move_group.retime_trajectory(self.robot_commander.get_current_state(), plan,
                                                     self.max_velocity_scaling_factor)
            rospy.loginfo("Planning time to move to pick: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.execute(plan, wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
            else:
                rospy.logerr("Failed to execute pick")
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Pick joint angles:")
                self.get_joint_angles()

            # --------------------- retract pick --------------------------------------
            retract_pick_goal = copy.deepcopy(approach_pick_goal)
            waypoints[0] = retract_pick_goal
            self.move_group.set_start_state_to_current_state()
            (plan, fraction) = self.move_group.compute_cartesian_path(waypoints, 0.05, 0)
            plan = self.move_group.retime_trajectory(self.robot_commander.get_current_state(), plan,
                                                     self.max_velocity_scaling_factor)
            rospy.loginfo("Planning time to move to retract pick: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.execute(plan, wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
            else:
                rospy.logerr("Failed to execute retract pick")
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Pick Retract joint angles:")
                self.get_joint_angles()

            # --------------------- approach place --------------------------------------
            approach_place_goal = copy.deepcopy(approach_pick_goal)
            approach_place_goal.position.y = approach_pick_goal.position.y - 0.6
            # self.move_group.clear_pose_targets()
            # self.move_group.set_pose_target(approach_place_goal)
            # -------------------
            # This move is hard coded with joint angles to remove the fact that
            # MoveIt does not always find the same final joint pose
            # -------------------
            approach_place_joint_angles = [-0.70778, 0.12450, -0.62438, 0.00071, -0.82112, 0.70677]
            self.move_group.set_joint_value_target(approach_place_joint_angles)
            self.move_group.set_start_state_to_current_state()

            plan = self.move_group.plan()
            rospy.loginfo("Planning time to move to approach place: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.go(wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
                rospy.loginfo("Made it to approach place with the following joint angle error:")
                self.diff_joint_angles(approach_place_joint_angles, self.get_joint_angles(print_joint_angles=False))
            else:
                rospy.logerr("Failed to execute approach place")
            rospy.loginfo("Execute time: %s", (time.time() - start_time_))
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Approach Place joint angles:")
                self.get_joint_angles()

            # --------------------- place --------------------------------------
            place_goal = copy.deepcopy(approach_place_goal)
            place_goal.position.z = approach_place_goal.position.z - 0.2
            waypoints[0] = place_goal
            self.move_group.set_start_state_to_current_state()
            (plan, fraction) = self.move_group.compute_cartesian_path(waypoints, 0.05, 0)
            plan = self.move_group.retime_trajectory(self.robot_commander.get_current_state(), plan,
                                                     self.max_velocity_scaling_factor)
            rospy.loginfo("Planning time to move to place: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.execute(plan, wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
            else:
                rospy.logerr("Failed to execute place")
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Place joint angles:")
                self.get_joint_angles()

            # --------------------- retract place --------------------------------------
            retract_place_goal = copy.deepcopy(approach_place_goal)
            waypoints[0] = retract_place_goal
            self.move_group.set_start_state_to_current_state()
            (plan, fraction) = self.move_group.compute_cartesian_path(waypoints, 0.05, 0)
            plan = self.move_group.retime_trajectory(self.robot_commander.get_current_state(), plan,
                                                     self.max_velocity_scaling_factor)
            rospy.loginfo("Planning time to move to retract place: %s", (time.time() - start_time_))

            if self.wait_for_user:
                raw_input("Press Enter to move arm...")
            start_time_ = time.time()
            success = self.move_group.execute(plan, wait=True)
            self.move_group.stop()
            if success:
                rospy.loginfo("Execute time: %s", (time.time() - start_time_))
            else:
                rospy.logerr("Failed to execute retract place")
            start_time_ = time.time()
            time.sleep(self.dwell_time_between_moves)
            if self.print_joint_angles:
                rospy.loginfo("Retract Pick joint angles:")
                self.get_joint_angles()


def main():
    try:
        rospy.init_node("y_dir_pnp_move", anonymous=True, disable_signals=True)
        r = RobotManager()
        r.move_arm()
    except KeyboardInterrupt:
        rospy.signal_shutdown("KeyboardInterrupt")
        raise


if __name__ == '__main__': main()
